#!/usr/bin/env python
#encoding:utf-8
#author:dbr/Ben
#project:tvdb_api
#repository:http://github.com/dbr/tvdb_api
#license:Creative Commons GNU GPL v2
# (http://creativecommons.org/licenses/GPL/2.0/)

"""
urllib2 caching handler
Modified from http://code.activestate.com/recipes/491261/
"""
from __future__ import with_statement

__author__ = "dbr/Ben"
__version__ = "1.4"

import os
import time
import errno
import httplib
import urllib2
import StringIO
from hashlib import md5
from threading import RLock

cache_lock = RLock()

def locked_function(origfunc):
    """Decorator to execute function under lock"""
    def wrapped(*args, **kwargs):
        cache_lock.acquire()
        try:
            return origfunc(*args, **kwargs)
        finally:
            cache_lock.release()
    return wrapped

def calculate_cache_path(cache_location, url):
    """Checks if [cache_location]/[hash_of_url].headers and .body exist
    """
    thumb = md5(url).hexdigest()
    header = os.path.join(cache_location, thumb + ".headers")
    body = os.path.join(cache_location, thumb + ".body")
    return header, body

def check_cache_time(path, max_age):
    """Checks if a file has been created/modified in the [last max_age] seconds.
    False means the file is too old (or doesn't exist), True means it is
    up-to-date and valid"""
    if not os.path.isfile(path):
        return False
    cache_modified_time = os.stat(path).st_mtime
    time_now = time.time()
    if cache_modified_time < time_now - max_age:
        # Cache is old
        return False
    else:
        return True

@locked_function
def exists_in_cache(cache_location, url, max_age):
    """Returns if header AND body cache file exist (and are up-to-date)"""
    hpath, bpath = calculate_cache_path(cache_location, url)
    if os.path.exists(hpath) and os.path.exists(bpath):
        return(
            check_cache_time(hpath, max_age)
            and check_cache_time(bpath, max_age)
        )
    else:
        # File does not exist
        return False

@locked_function
def store_in_cache(cache_location, url, response):
    """Tries to store response in cache."""
    hpath, bpath = calculate_cache_path(cache_location, url)
    try:
        outf = open(hpath, "w")
        headers = str(response.info())
        outf.write(headers)
        outf.close()

        outf = open(bpath, "w")
        outf.write(response.read())
        outf.close()
    except IOError:
        return True
    else:
        return False

class CacheHandler(urllib2.BaseHandler):
    """Stores responses in a persistant on-disk cache.

    If a subsequent GET request is made for the same URL, the stored
    response is returned, saving time, resources and bandwidth
    """
    @locked_function
    def __init__(self, cache_location, max_age = 21600):
        """The location of the cache directory"""
        self.max_age = max_age
        self.cache_location = cache_location
        if not os.path.exists(self.cache_location):
            try:
                os.mkdir(self.cache_location)
            except OSError, e:
                if e.errno == errno.EEXIST and os.path.isdir(self.cache_location):
                    # File exists, and it's a directory,
                    # another process beat us to creating this dir, that's OK.
                    pass
                else:
                    # Our target dir is already a file, or different error,
                    # relay the error!
                    raise

    def default_open(self, request):
        """Handles GET requests, if the response is cached it returns it
        """
        if request.get_method() is not "GET":
            return None # let the next handler try to handle the request

        if exists_in_cache(
            self.cache_location, request.get_full_url(), self.max_age
        ):
            return CachedResponse(
                self.cache_location,
                request.get_full_url(),
                set_cache_header = True
            )
        else:
            return None

    def http_response(self, request, response):
        """Gets a HTTP response, if it was a GET request and the status code
        starts with 2 (200 OK etc) it caches it and returns a CachedResponse
        """
        if (request.get_method() == "GET"
            and str(response.code).startswith("2")
        ):
            if 'x-local-cache' not in response.info():
                # Response is not cached
                set_cache_header = store_in_cache(
                    self.cache_location,
                    request.get_full_url(),
                    response
                )
            else:
                set_cache_header = True
            #end if x-cache in response

            return CachedResponse(
                self.cache_location,
                request.get_full_url(),
                set_cache_header = set_cache_header
            )
        else:
            return response

class CachedResponse(StringIO.StringIO):
    """An urllib2.response-like object for cached responses.

    To determine if a response is cached or coming directly from
    the network, check the x-local-cache header rather than the object type.
    """

    @locked_function
    def __init__(self, cache_location, url, set_cache_header=True):
        self.cache_location = cache_location
        hpath, bpath = calculate_cache_path(cache_location, url)

        StringIO.StringIO.__init__(self, file(bpath).read())

        self.url     = url
        self.code    = 200
        self.msg     = "OK"
        headerbuf = file(hpath).read()
        if set_cache_header:
            headerbuf += "x-local-cache: %s\r\n" % (bpath)
        self.headers = httplib.HTTPMessage(StringIO.StringIO(headerbuf))

    def info(self):
        """Returns headers
        """
        return self.headers

    def geturl(self):
        """Returns original URL
        """
        return self.url

    @locked_function
    def recache(self):
        new_request = urllib2.urlopen(self.url)
        set_cache_header = store_in_cache(
            self.cache_location,
            new_request.url,
            new_request
        )
        CachedResponse.__init__(self, self.cache_location, self.url, True)


if __name__ == "__main__":
    def main():
        """Quick test/example of CacheHandler"""
        opener = urllib2.build_opener(CacheHandler("/tmp/"))
        response = opener.open("http://google.com")
        print response.headers
        print "Response:", response.read()

        response.recache()
        print response.headers
        print "After recache:", response.read()

        # Test usage in threads
        from threading import Thread
        class CacheThreadTest(Thread):
            lastdata = None
            def run(self):
                req = opener.open("http://google.com")
                newdata = req.read()
                if self.lastdata is None:
                    self.lastdata = newdata
                assert self.lastdata == newdata, "Data was not consistent, uhoh"
                req.recache()
        threads = [CacheThreadTest() for x in range(50)]
        print "Starting threads"
        [t.start() for t in threads]
        print "..done"
        print "Joining threads"
        [t.join() for t in threads]
        print "..done"
    main()



#!/usr/bin/env python
#encoding:utf-8
#author:dbr/Ben
#project:tvdb_api
#repository:http://github.com/dbr/tvdb_api
#license:Creative Commons GNU GPL v2
# (http://creativecommons.org/licenses/GPL/2.0/)

"""Simple-to-use Python interface to The TVDB's API (www.thetvdb.com)

Example usage:

>>> from tvdb_api import Tvdb
>>> t = Tvdb()
>>> t['Lost'][4][11]['episodename']
u'Cabin Fever'
"""
__author__ = "dbr/Ben"
__version__ = "1.4"

import os
import sys
import urllib
import urllib2
import tempfile
import logging

try:
    import xml.etree.cElementTree as ElementTree
except ImportError:
    import xml.etree.ElementTree as ElementTree

from cache import CacheHandler

from tvdb_ui import BaseUI, ConsoleUI
from tvdb_exceptions import (tvdb_error, tvdb_userabort, tvdb_shownotfound,
    tvdb_seasonnotfound, tvdb_episodenotfound, tvdb_attributenotfound)

class ShowContainer(dict):
    """Simple dict that holds a series of Show instances
    """
    pass


class Show(dict):
    """Holds a dict of seasons, and show data.
    """
    def __init__(self):
        dict.__init__(self)
        self.data = {}

    def __repr__(self):
        return "<Show %s (containing %s seasons)>" % (
            self.data.get(u'seriesname', 'instance'),
            len(self)
        )

    def __getitem__(self, key):
        if key in self:
            # Key is an episode, return it
            return dict.__getitem__(self, key)

        if key in self.data:
            # Non-numeric request is for show-data
            return dict.__getitem__(self.data, key)

        # Data wasn't found, raise appropriate error
        if isinstance(key, int) or key.isdigit():
            # Episode number x was not found
            raise tvdb_seasonnotfound("Could not find season %s" % (repr(key)))
        else:
            # If it's not numeric, it must be an attribute name, which
            # doesn't exist, so attribute error.
            raise tvdb_attributenotfound("Cannot find attribute %s" % (repr(key)))

    def search(self, term = None, key = None):
        """
        Search all episodes in show. Can search all data, or a specific key (for
        example, episodename)

        Always returns an array (can be empty). First index contains the first
        match, and so on.

        Each array index is an Episode() instance, so doing
        search_results[0]['episodename'] will retrieve the episode name of the
        first match.

        Search terms are converted to lower case (unicode) strings.

        # Examples
        
        These examples assume t is an instance of Tvdb():
        
        >>> t = Tvdb()
        >>>

        To search for all episodes of Scrubs with a bit of data
        containing "my first day":

        >>> t['Scrubs'].search("my first day")
        [<Episode 01x01 - My First Day>]
        >>>

        Search for "My Name Is Earl" episode named "Faked His Own Death":

        >>> t['My Name Is Earl'].search('Faked His Own Death', key = 'episodename')
        [<Episode 01x04 - Faked His Own Death>]
        >>>

        To search Scrubs for all episodes with "mentor" in the episode name:

        >>> t['scrubs'].search('mentor', key = 'episodename')
        [<Episode 01x02 - My Mentor>, <Episode 03x15 - My Tormented Mentor>]
        >>>

        # Using search results

        >>> results = t['Scrubs'].search("my first")
        >>> print results[0]['episodename']
        My First Day
        >>> for x in results: print x['episodename']
        My First Day
        My First Step
        My First Kill
        >>>
        """
        results = []
        for cur_season in self.values():
            searchresult = cur_season.search(term = term, key = key)
            if len(searchresult) != 0:
                results.extend(searchresult)
        #end for cur_season
        return results


class Season(dict):
    def __repr__(self):
        return "<Season instance (containing %s episodes)>" % (
            len(self.keys())
        )

    def __getitem__(self, episode_number):
        if episode_number not in self:
            raise tvdb_episodenotfound("Could not find episode %s" % (repr(episode_number)))
        else:
            return dict.__getitem__(self, episode_number)

    def search(self, term = None, key = None):
        """Search all episodes in season, returns a list of matching Episode
        instances.

        >>> t = Tvdb()
        >>> t['scrubs'][1].search('first day')
        [<Episode 01x01 - My First Day>]
        >>>

        See Show.search documentation for further information on search
        """
        results = []
        for ep in self.values():
            searchresult = ep.search(term = term, key = key)
            if searchresult is not None:
                results.append(
                    searchresult
                )
        return results


class Episode(dict):
    def __repr__(self):
        seasno = int(self.get(u'seasonnumber', 0))
        epno = int(self.get(u'episodenumber', 0))
        epname = self.get(u'episodename')
        if epname is not None:
            return "<Episode %02dx%02d - %s>" % (seasno, epno, epname)
        else:
            return "<Episode %02dx%02d>" % (seasno, epno)

    def __getitem__(self, key):
        try:
            return dict.__getitem__(self, key)
        except KeyError:
            raise tvdb_attributenotfound("Cannot find attribute %s" % (repr(key)))

    def search(self, term = None, key = None):
        """Search episode data for term, if it matches, return the Episode (self).
        The key parameter can be used to limit the search to a specific element,
        for example, episodename.
        
        This primarily for use use by Show.search and Season.search. See
        Show.search for further information on search

        Simple example:

        >>> e = Episode()
        >>> e['episodename'] = "An Example"
        >>> e.search("examp")
        <Episode 00x00 - An Example>
        >>>

        Limiting by key:

        >>> e.search("examp", key = "episodename")
        <Episode 00x00 - An Example>
        >>>
        """
        if term == None:
            raise TypeError("must supply string to search for (contents)")

        term = unicode(term).lower()
        for cur_key, cur_value in self.items():
            cur_key, cur_value = unicode(cur_key).lower(), unicode(cur_value).lower()
            if key is not None and cur_key != key:
                # Do not search this key
                continue
            if cur_value.find( unicode(term).lower() ) > -1:
                return self
            #end if cur_value.find()
        #end for cur_key, cur_value


class Actors(list):
    """Holds all Actor instances for a show
    """
    pass


class Actor(dict):
    """Represents a single actor. Should contain..

    id,
    image,
    name,
    role,
    sortorder
    """
    def __repr__(self):
        return "<Actor \"%s\">" % (self.get("name"))


class Tvdb:
    """Create easy-to-use interface to name of season/episode name
    >>> t = Tvdb()
    >>> t['Scrubs'][1][24]['episodename']
    u'My Last Day'
    """
    def __init__(self,
                interactive = False,
                select_first = False,
                debug = False,
                cache = True,
                banners = False,
                actors = False,
                custom_ui = None,
                language = None,
                search_all_languages = False,
                apikey = None):
        """interactive (True/False):
            When True, uses built-in console UI is used to select the correct show.
            When False, the first search result is used.

        select_first (True/False):
            Automatically selects the first series search result (rather
            than showing the user a list of more than one series).
            Is overridden by interactive = False, or specifying a custom_ui

        debug (True/False):
             shows verbose debugging information

        cache (True/False/str/unicode):
            Retrieved XML are persisted to to disc. If true, stores in tvdb_api
            folder under your systems TEMP_DIR, if set to str/unicode instance it
            will use this as the cache location. If False, disables caching.

        banners (True/False):
            Retrieves the banners for a show. These are accessed
            via the _banners key of a Show(), for example:

            >>> Tvdb(banners=True)['scrubs']['_banners'].keys()
            ['fanart', 'poster', 'series', 'season']

        actors (True/False):
            Retrieves a list of the actors for a show. These are accessed
            via the _actors key of a Show(), for example:

            >>> t = Tvdb(actors=True)
            >>> t['scrubs']['_actors'][0]['name']
            u'Zach Braff'

        custom_ui (tvdb_ui.BaseUI subclass):
            A callable subclass of tvdb_ui.BaseUI (overrides interactive option)

        language (2 character language abbreviation):
            The language of the returned data. Is also the language search
            uses. Default is "en" (English). For full list, run..

            >>> Tvdb().config['valid_languages'] #doctest: +ELLIPSIS
            ['da', 'fi', 'nl', ...]

        search_all_languages (True/False):
            By default, Tvdb will only search in the language specified using
            the language option. When this is True, it will search for the
            show in and language
        
        apikey (str/unicode):
            Override the default thetvdb.com API key. By default it will use
            tvdb_api's own key (fine for small scripts), but you can use your
            own key if desired - this is recommended if you are embedding
            tvdb_api in a larger application)
            See http://thetvdb.com/?tab=apiregister to get your own key
        """
        self.shows = ShowContainer() # Holds all Show classes
        self.corrections = {} # Holds show-name to show_id mapping

        self.config = {}

        if apikey is not None:
            self.config['apikey'] = apikey
        else:
            self.config['apikey'] = "0629B785CE550C8D" # tvdb_api's API key

        self.config['debug_enabled'] = debug # show debugging messages

        self.config['custom_ui'] = custom_ui

        self.config['interactive'] = interactive # prompt for correct series?

        self.config['select_first'] = select_first

        self.config['search_all_languages'] = search_all_languages

        if cache is True:
            self.config['cache_enabled'] = True
            self.config['cache_location'] = self._getTempDir()
        elif isinstance(cache, basestring):
            self.config['cache_enabled'] = True
            self.config['cache_location'] = cache
        else:
            self.config['cache_enabled'] = False

        if self.config['cache_enabled']:
            self.urlopener = urllib2.build_opener(
                CacheHandler(self.config['cache_location'])
            )
        else:
            self.urlopener = urllib2.build_opener()

        self.config['banners_enabled'] = banners
        self.config['actors_enabled'] = actors

        self.log = self._initLogger() # Setups the logger (self.log.debug() etc)

        # List of language from http://www.thetvdb.com/api/0629B785CE550C8D/languages.xml
        # Hard-coded here as it is realtively static, and saves another HTTP request, as
        # recommended on http://thetvdb.com/wiki/index.php/API:languages.xml
        self.config['valid_languages'] = [
            "da", "fi", "nl", "de", "it", "es", "fr","pl", "hu","el","tr",
            "ru","he","ja","pt","zh","cs","sl", "hr","ko","en","sv","no"
        ]

        # thetvdb.com should be based around numeric language codes,
        # but to link to a series like http://thetvdb.com/?tab=series&id=79349&lid=16
        # requires the language ID, thus this mapping is required (mainly
        # for usage in tvdb_ui - internally tvdb_api will use the language abbreviations)
        self.config['langabbv_to_id'] = {'el': 20, 'en': 7, 'zh': 27,
        'it': 15, 'cs': 28, 'es': 16, 'ru': 22, 'nl': 13, 'pt': 26, 'no': 9,
        'tr': 21, 'pl': 18, 'fr': 17, 'hr': 31, 'de': 14, 'da': 10, 'fi': 11,
        'hu': 19, 'ja': 25, 'he': 24, 'ko': 32, 'sv': 8, 'sl': 30}

        if language is None:
            self.config['language'] = None
        else:
            if language not in self.config['valid_languages']:
                raise ValueError("Invalid language %s, options are: %s" % (
                    language, self.config['valid_languages']
                ))
            else:
                self.config['language'] = language

        # The following url_ configs are based of the
        # http://thetvdb.com/wiki/index.php/Programmers_API
        self.config['base_url'] = "http://www.thetvdb.com"

        if self.config['search_all_languages']:
            self.config['url_getSeries'] = "%(base_url)s/api/GetSeries.php?seriesname=%%s&language=all" % self.config
        else:
            self.config['url_getSeries'] = "%(base_url)s/api/GetSeries.php?seriesname=%%s&language=%(language)s" % self.config

        self.config['url_epInfo'] = "%(base_url)s/api/%(apikey)s/series/%%s/all/%%s.xml" % self.config

        self.config['url_seriesInfo'] = "%(base_url)s/api/%(apikey)s/series/%%s/%%s.xml" % self.config
        self.config['url_actorsInfo'] = "%(base_url)s/api/%(apikey)s/series/%%s/actors.xml" % self.config

        self.config['url_seriesBanner'] = "%(base_url)s/api/%(apikey)s/series/%%s/banners.xml" % self.config
        self.config['url_artworkPrefix'] = "%(base_url)s/banners/%%s" % self.config

    #end __init__

    def _initLogger(self):
        """Setups a logger using the logging module, returns a log object
        """
        logger = logging.getLogger("tvdb")
        formatter = logging.Formatter('%(asctime)s) %(levelname)s %(message)s')

        hdlr = logging.StreamHandler(sys.stdout)

        hdlr.setFormatter(formatter)
        logger.addHandler(hdlr)

        if self.config['debug_enabled']:
            logger.setLevel(logging.DEBUG)
        else:
            logger.setLevel(logging.WARNING)
        return logger
    #end initLogger

    def _getTempDir(self):
        """Returns the [system temp dir]/tvdb_api
        """
        return os.path.join(tempfile.gettempdir(), "tvdb_api")

    def _loadUrl(self, url, recache = False):
        try:
            self.log.debug("Retrieving URL %s" % url)
            resp = self.urlopener.open(url)
            if 'x-local-cache' in resp.headers:
                self.log.debug("URL %s was cached in %s" % (
                    url,
                    resp.headers['x-local-cache'])
                )
                if recache:
                    self.log.debug("Attempting to recache %s" % url)
                    resp.recache()
        except urllib2.URLError, errormsg:
            raise tvdb_error("Could not connect to server: %s" % (errormsg))
        #end try

        return resp.read()

    def _getetsrc(self, url):
        """Loads a URL sing caching, returns an ElementTree of the source
        """
        src = self._loadUrl(url)
        try:
            return ElementTree.fromstring(src)
        except SyntaxError:
            src = self._loadUrl(url, recache=True)
            try:
                return ElementTree.fromstring(src)
            except SyntaxError, exceptionmsg:
                errormsg = "There was an error with the XML retrieved from thetvdb.com:\n%s" % (
                    exceptionmsg
                )

                if self.config['cache_enabled']:
                    errormsg += "\nFirst try emptying the cache folder at..\n%s" % (
                        self.config['cache_location']
                    )

                errormsg += "\nIf this does not resolve the issue, please try again later. If the error persists, report a bug on"
                errormsg += "\nhttp://dbr.lighthouseapp.com/projects/13342-tvdb_api/overview\n"
                raise tvdb_error(errormsg)
    #end _getetsrc

    def _setItem(self, sid, seas, ep, attrib, value):
        """Creates a new episode, creating Show(), Season() and
        Episode()s as required. Called by _getShowData to populate show

        Since the nice-to-use tvdb[1][24]['name] interface
        makes it impossible to do tvdb[1][24]['name] = "name"
        and still be capable of checking if an episode exists
        so we can raise tvdb_shownotfound, we have a slightly
        less pretty method of setting items.. but since the API
        is supposed to be read-only, this is the best way to
        do it!
        The problem is that calling tvdb[1][24]['episodename'] = "name"
        calls __getitem__ on tvdb[1], there is no way to check if
        tvdb.__dict__ should have a key "1" before we auto-create it
        """
        if sid not in self.shows:
            self.shows[sid] = Show()
        if seas not in self.shows[sid]:
            self.shows[sid][seas] = Season()
        if ep not in self.shows[sid][seas]:
            self.shows[sid][seas][ep] = Episode()
        self.shows[sid][seas][ep][attrib] = value
    #end _set_item

    def _setShowData(self, sid, key, value):
        """Sets self.shows[sid] to a new Show instance, or sets the data
        """
        if sid not in self.shows:
            self.shows[sid] = Show()
        self.shows[sid].data[key] = value

    def _cleanData(self, data):
        """Cleans up strings returned by TheTVDB.com

        Issues corrected:
        - Replaces &amp; with &
        - Trailing whitespace
        """
        data = data.replace(u"&amp;", u"&")
        data = data.strip()
        return data
    #end _cleanData

    def _getSeries(self, series):
        """This searches TheTVDB.com for the series name,
        If a custom_ui UI is configured, it uses this to select the correct
        series. If not, and interactive == True, ConsoleUI is used, if not
        BaseUI is used to select the first result.
        """
        series = urllib.quote(series.encode("utf-8"))
        self.log.debug("Searching for show %s" % series)
        seriesEt = self._getetsrc(self.config['url_getSeries'] % (series))
        allSeries = []
        for series in seriesEt:
            result = dict((k.tag.lower(), k.text) for k in series.getchildren())
            result['lid'] = self.config['langabbv_to_id'][result['language']]
            self.log.debug('Found series %(seriesname)s' % result)
            allSeries.append(result)
        #end for series

        if len(allSeries) == 0:
            self.log.debug('Series result returned zero')
            raise tvdb_shownotfound("Show-name search returned zero results (cannot find show on TVDB)")

        if self.config['custom_ui'] is not None:
            self.log.debug("Using custom UI %s" % (repr(self.config['custom_ui'])))
            ui = self.config['custom_ui'](config = self.config, log = self.log)
        else:
            if not self.config['interactive']:
                self.log.debug('Auto-selecting first search result using BaseUI')
                ui = BaseUI(config = self.config, log = self.log)
            else:
                self.log.debug('Interactively selecting show using ConsoleUI')
                ui = ConsoleUI(config = self.config, log = self.log)
            #end if config['interactive]
        #end if custom_ui != None

        return ui.selectSeries(allSeries)

    #end _getSeries

    def _parseBanners(self, sid):
        """Parses banners XML, from
        http://www.thetvdb.com/api/[APIKEY]/series/[SERIES ID]/banners.xml

        Banners are retrieved using t['show name]['_banners'], for example:

        >>> t = Tvdb(banners = True)
        >>> t['scrubs']['_banners'].keys()
        ['fanart', 'poster', 'series', 'season']
        >>> t['scrubs']['_banners']['poster']['680x1000']['35308']['_bannerpath']
        'http://www.thetvdb.com/banners/posters/76156-2.jpg'
        >>>

        Any key starting with an underscore has been processed (not the raw
        data from the XML)

        This interface will be improved in future versions.
        """
        self.log.debug('Getting season banners for %s' % (sid))
        bannersEt = self._getetsrc( self.config['url_seriesBanner'] % (sid) )
        banners = {}
        for cur_banner in bannersEt.findall('Banner'):
            bid = cur_banner.find('id').text
            btype = cur_banner.find('BannerType')
            btype2 = cur_banner.find('BannerType2')
            if btype is None or btype2 is None:
                continue
            btype, btype2 = btype.text, btype2.text
            if not btype in banners:
                banners[btype] = {}
            if not btype2 in banners[btype]:
                banners[btype][btype2] = {}
            if not bid in banners[btype][btype2]:
                banners[btype][btype2][bid] = {}

            for cur_element in cur_banner.getchildren():
                tag = cur_element.tag.lower()
                value = cur_element.text
                if tag is None or value is None:
                    continue
                tag, value = tag.lower(), value.lower()
                banners[btype][btype2][bid][tag] = value

            for k, v in banners[btype][btype2][bid].items():
                if k.endswith("path"):
                    new_key = "_%s" % (k)
                    self.log.debug("Transforming %s to %s" % (k, new_key))
                    new_url = self.config['url_artworkPrefix'] % (v)
                    banners[btype][btype2][bid][new_key] = new_url

        self._setShowData(sid, "_banners", banners)

    def _parseActors(self, sid):
        """Parsers actors XML, from
        http://www.thetvdb.com/api/[APIKEY]/series/[SERIES ID]/actors.xml

        Actors are retrieved using t['show name]['_actors'], for example:

        >>> t = Tvdb(actors = True)
        >>> actors = t['scrubs']['_actors']
        >>> type(actors)
        <class 'tvdb_api.Actors'>
        >>> type(actors[0])
        <class 'tvdb_api.Actor'>
        >>> actors[0]
        <Actor "Zach Braff">
        >>> sorted(actors[0].keys())
        ['id', 'image', 'name', 'role', 'sortorder']
        >>> actors[0]['name']
        u'Zach Braff'
        >>> actors[0]['image']
        'http://www.thetvdb.com/banners/actors/43640.jpg'

        Any key starting with an underscore has been processed (not the raw
        data from the XML)
        """
        self.log.debug("Getting actors for %s" % (sid))
        actorsEt = self._getetsrc(self.config['url_actorsInfo'] % (sid))

        cur_actors = Actors()
        for curActorItem in actorsEt.findall("Actor"):
            curActor = Actor()
            for curInfo in curActorItem:
                tag = curInfo.tag.lower()
                value = curInfo.text
                if value is not None:
                    if tag == "image":
                        value = self.config['url_artworkPrefix'] % (value)
                    else:
                        value = self._cleanData(value)
                curActor[tag] = value
            cur_actors.append(curActor)
        self._setShowData(sid, '_actors', cur_actors)

    def _getShowData(self, sid, language):
        """Takes a series ID, gets the epInfo URL and parses the TVDB
        XML file into the shows dict in layout:
        shows[series_id][season_number][episode_number]
        """

        if self.config['language'] is None:
            self.log.debug('Config language is none, using show language')
            getShowInLanguage = language
        else:
            self.log.debug(
                'Configured language %s override show language of %s' % (
                    self.config['language'],
                    language
                )
            )
            getShowInLanguage = self.config['language']

        # Parse show information
        self.log.debug('Getting all series data for %s' % (sid))
        seriesInfoEt = self._getetsrc(
            self.config['url_seriesInfo'] % (sid, getShowInLanguage)
        )
        for curInfo in seriesInfoEt.findall("Series")[0]:
            tag = curInfo.tag.lower()
            value = curInfo.text

            if value is not None:
                if tag in ['banner', 'fanart', 'poster']:
                    value = self.config['url_artworkPrefix'] % (value)
                else:
                    value = self._cleanData(value)

            self._setShowData(sid, tag, value)
        #end for series

        # Parse banners
        if self.config['banners_enabled']:
            self._parseBanners(sid)

        # Parse actors
        if self.config['actors_enabled']:
            self._parseActors(sid)

        # Parse episode data
        self.log.debug('Getting all episodes of %s' % (sid))
        epsEt = self._getetsrc( self.config['url_epInfo'] % (sid, language) )

        for cur_ep in epsEt.findall("Episode"):
            seas_no = int(cur_ep.find('SeasonNumber').text)
            ep_no = int(cur_ep.find('EpisodeNumber').text)
            for cur_item in cur_ep.getchildren():
                tag = cur_item.tag.lower()
                value = cur_item.text
                if value is not None:
                    if tag == 'filename':
                        value = self.config['url_artworkPrefix'] % (value)
                    else:
                        value = self._cleanData(value)
                self._setItem(sid, seas_no, ep_no, tag, value)
        #end for cur_ep
    #end _geEps

    def _nameToSid(self, name):
        """Takes show name, returns the correct series ID (if the show has
        already been grabbed), or grabs all episodes and returns
        the correct SID.
        """
        if name in self.corrections:
            self.log.debug('Correcting %s to %s' % (name, self.corrections[name]) )
            sid = self.corrections[name]
        else:
            self.log.debug('Getting show %s' % (name))
            selected_series = self._getSeries( name )
            sname, sid = selected_series['seriesname'], selected_series['id']
            self.log.debug('Got %(seriesname)s, id %(id)s' % selected_series)

            self.corrections[name] = sid
            self._getShowData(selected_series['id'], selected_series['language'])
        #end if name in self.corrections
        return sid
    #end _nameToSid

    def __getitem__(self, key):
        """Handles tvdb_instance['seriesname'] calls.
        The dict index should be the show id
        """
        if isinstance(key, (int, long)):
            # Item is integer, treat as show id
            if key not in self.shows:
                self._getShowData(key, self.config['language'])
            return self.shows[key]
        
        key = key.lower() # make key lower case
        sid = self._nameToSid(key)
        self.log.debug('Got series id %s' % (sid))
        return self.shows[sid]
    #end __getitem__

    def __repr__(self):
        return str(self.shows)
    #end __repr__
#end Tvdb

def main():
    """Simple example of using tvdb_api - it just
    grabs an episode name interactively.
    """
    tvdb_instance = Tvdb(interactive=True, debug=True, cache=False)
    print tvdb_instance['Lost']['seriesname']
    print tvdb_instance['Lost'][1][4]['episodename']

if __name__ == '__main__':
    main()



#!/usr/bin/env python
#encoding:utf-8
#author:tuxtof
#project:tvtags
#license:Creative Commons GNU GPL v2
# (http://creativecommons.org/licenses/GPL/2.0/)
"""
tvtags.py
Automatic TV episode tagger.
Uses data from www.thetvdb.com via tvdb_api,

thanks goes to:
dbr/Ben - http://github.com/dbr - for python help, tvdb_api and tvnamer.py
Rodney - http://kerstetter.net - for MP4Tagger help
ccjensen/Chris - for mp4tvtagger (source of this code)
"""

__author__ = "tuxtof"

import os
import sys
import re
import glob
import tempfile
import urllib
from unicodedata import normalize

from optparse import OptionParser
import subprocess

from tvdb_api.tvdb_api import Tvdb
from tvdb_api.tvdb_exceptions import (tvdb_error, tvdb_userabort, tvdb_shownotfound,
    tvdb_seasonnotfound, tvdb_episodenotfound, tvdb_attributenotfound)

import progressbar

class Program:
    """docstring for Program"""
    def __init__(self, opts, filePath, fileName):
        if opts.verbose:
            print "Connecting to the TVDB... "
        #end if verbose
        tvDebug = False
        if opts.verbose > 1:
            tvDebug = True
        self.tvdb = Tvdb(debug = tvDebug, interactive = opts.interactive, banners = True, language = "fr")
        self.MP4Tagger = "MP4Tagger"
        self.filePath = filePath
        self.fileName = fileName
    #end def __init__
#end class Program


class Series:
    """docstring for Series"""
    artworkFileName = ""
    
    def __init__(self, opts, program, series, seasonNumber):
        #get show specific meta data
        if opts.verbose > 0:
            print "Retrieving Show Information... "
        #end if verbose
        self.seriesName = getShowSpecificInfo(opts, program.tvdb, series, 'seriesname')
        
        self.actorsUnsplit = getShowSpecificInfo(opts, program.tvdb, self.seriesName, 'actors')
        self.actors = self.actorsUnsplit.split('|')
        
        self.contentRating = getShowSpecificInfo(opts, program.tvdb, self.seriesName, 'contentrating')
        #self.firstaired  = getShowSpecificInfo(opts, program.tvdb, self.seriesName, 'firstaired') #currently not used for anything
        
        self.genresUnsplit  = getShowSpecificInfo(opts, program.tvdb, self.seriesName, 'genre')
        self.genres = self.genresUnsplit.split('|')
        
        self.network  = getShowSpecificInfo(opts, program.tvdb, self.seriesName, 'network')
        #seriesOverview  = getShowSpecificInfo(opts, program.tvdb, self.seriesName, 'overview') #currently not used for anything
        
        self.seasonNumber = int(seasonNumber)
    #end def __init__
#end class Series


class Episode:
    """docstring for Episode"""
    def __init__(self, verbose, program, series, seasonNumberEpisode, episodeNumber):
        self.fileName = program.fileName
        self.filePath = program.filePath
        
        
        #pattern = re.compile('[\D]+')
        
        
        
        #Parse the file name for information: 1x01 - Pilot.mp4
        (fileBaseName, self.fileExtension) = os.path.splitext(self.fileName)
        #(seasonNumberEpisode, episodeNumber, tail) = pattern.split(fileBaseName,2)
        
        #check if filename was of correct format, else set it to an incorrect value
        if len(seasonNumberEpisode) > 0:
            self.seasonNumberEpisode = int(seasonNumberEpisode)
        else:
            self.seasonNumberEpisode = 9999
        #end if len(seasonNumberEpisode)
        
        self.episodeNumber = int(episodeNumber)
        
        #get other episode specific meta data
        self.episodeName = getEpisodeSpecificInfo(verbose, program, series, self.episodeNumber, 'episodename')
        self.firstAired = getEpisodeSpecificInfo(verbose, program, series, self.episodeNumber, 'firstaired') + "T09:00:00Z"
        
        self.guestStarsUnsplit = getEpisodeSpecificInfo(verbose, program, series, self.episodeNumber, 'gueststars')
        # if self.guestStarsUnsplit:
        #   self.guestStars = self.guestStarsUnsplit.split('|')
        # else:
        #   self.guestStars = ""
        
        self.directorsUnsplit = getEpisodeSpecificInfo(verbose, program, series, self.episodeNumber, 'director')
        self.directors = self.directorsUnsplit.split('|')
        
        self.writersUnsplit = getEpisodeSpecificInfo(verbose, program, series, self.episodeNumber, 'writer')
        self.writers = self.writersUnsplit.split('|')
        
        self.productionCode = getEpisodeSpecificInfo(verbose, program, series, self.episodeNumber, 'productioncode')
        self.overview = getEpisodeSpecificInfo(verbose, program, series, self.episodeNumber, 'overview')
    #end def __init__
#end class Episode


def whichBin(execName):
    "Try to find binary file"
    for path in os.environ["PATH"].split(":"):
        if os.path.exists(os.path.join(path, execName)):
            return 1


def openurl(urls):
    "docstring for openurl"
    for url in urls:
        if len(url) > 0:
            os.popen("open \"%s\"" % url)
        #end if len
    #end for url
    return
#end openurl


def correctFileName(verbose, program, series, episode):
    """docstring for correctFilename"""
    #Correct file name if incorrect
    if episode.fileName != "%s.%01dx%02d%s" % (series.seriesName.encode("utf-8").replace('/', "-"), series.seasonNumber, episode.episodeNumber, episode.fileExtension):
        newFileName = "%s.%01dx%02d%s" % (series.seriesName.encode("utf-8").replace('/', "-"),series.seasonNumber, episode.episodeNumber,  episode.fileExtension)
        renameCmd = "mv -n \"%s/%s\" \"%s/%s\"" % (episode.filePath, episode.fileName, episode.filePath, newFileName)
        os.popen(renameCmd)
        if verbose:
            print "Filename corrected from \"%s\" to \"%s\"" % (episode.fileName, newFileName)
        #end if verbose
        episode.fileName = newFileName
    else:
        if verbose:
            print "Filename \"%s\" already correct" % episode.fileName
        #end if verbose
    #end if fileName
#end correctFileName


def tagFile(opts, program, series, episode):
    """docstring for tagFile"""
    #setup tags for the MP4Tagger function
    if series.artworkFileName != "":
        addArtwork = " --artwork \"%s\"" % series.artworkFileName #the file we downloaded earlier
    else:
        addArtwork = ""
    #end if series.artworkFileName != ""
    addName = " --name=\"%ix%02i %s\"" % (series.seasonNumber, episode.episodeNumber, episode.episodeName)
    addStik = " --media_kind=\"TV Show\"" #set type to TV Show
    addArtist = " --artist \"%s\"" % series.seriesName
    addTitle =  " --tv_episode_id \"%s\"" % episode.episodeName
    addAlbum = " --album \"%s - Season %s\"" % (series.seriesName, series.seasonNumber)
    addGenre = " --genre \"%s\"" % series.genres[1] #cause first one is an empty string, and genre can only have one entry
    addAlbumArtist = " --album_artist \"%s\"" % series.seriesName
    addDescription = " --description \"%s\"" % episode.overview
    addLongDescription = " --long_description \"%s\"" % episode.overview
    addTVNetwork = " --tv_network \"%s\"" % series.network
    addTVShowName = " --tv_show \"%s\"" % series.seriesName
    addTVSeasonNum = " --tv_season \"%i\"" % series.seasonNumber
    addTVEpisodeNum = " --tv_episode_n \"%i\"" % episode.episodeNumber
    addDisk = " --disk_n \"%i\"" % series.seasonNumber
    addTracknum = " --track_n \"%i\"" % episode.episodeNumber
    addContentRating = " --rating \"%s\"" % series.contentRating
    addYear = " --release_date \"%s\"" % episode.firstAired
    addComment = " --comments \"tagged by tvtags\""
    if len(series.actors) > 0:
        addCast = " --cast \"%s\"" % ", ".join(series.actors)
    if len(episode.directors) > 0:
        addDirectors = " --director \"%s\"" % ", ".join(episode.directors)
    if len(episode.writers) > 0:
        addScreenWriters = " --screenwriters \"%s\"" % ", ".join(episode.writers)
    
    #Create the command line string
    tagArg = addName + addStik + addArtist + addTitle + addAlbum + addGenre + addAlbumArtist + addDescription \
    + addTVNetwork + addTVShowName +  addTVSeasonNum + addTVEpisodeNum + addDisk + addTracknum \
    + addContentRating  + addYear + addComment + addCast + addDirectors + addScreenWriters
    tagCmd = "\"" + program.MP4Tagger + "\" -i \"" + episode.filePath + "/" + episode.fileName + "\"" \
    + addArtwork + tagArg.encode("utf8")
    
    #run MP4Tagger using the arguments we have created
    if opts.verbose > 1:
        print tagCmd
    #end if debug
    
    execCmd = subprocess.Popen(tagCmd, shell=True)
    execCmd.wait()
    
    if opts.verbose:
        print "Tagged: " + episode.fileName
    #end if verbose

#end tagFile


def artwork(opts, interactive, program, series):
    cacheDir = os.path.join(tempfile.gettempdir(), "tvtags")
    
    if not os.path.exists(cacheDir):
        os.mkdir(cacheDir)
    
    try:
        potentialArtworkFileName = normalize('NFD', series.seriesName).encode("UTF-8") + " Season " + str(series.seasonNumber)
        for fullFileName in glob.glob(cacheDir + "/*.jpg"):
            (fileBaseName, fileExtension) = os.path.splitext(fullFileName)
            (filePath,fileName) = os.path.split(fileBaseName)
            
            if fileName == potentialArtworkFileName:
                if opts.verbose > 0:
                    print "Using Previously Downloaded Artwork: " + fileName
                #end if verbose
                series.artworkFileName = fullFileName
                return
            #end if fileBaseName
        #end for fileName
        
        tvdb = program.tvdb
        artworks = []
        
        if 'season' in tvdb[series.seriesName]['_banners']:
            if 'season' in tvdb[series.seriesName]['_banners']['season']:    
                for banner_id, banner_info in tvdb[series.seriesName]['_banners']['season']['season'].items():
                    if banner_info['season'] == str(series.seasonNumber):
                        artworks.append(banner_info['_bannerpath'])
        
        #check if we didn't find any artwork, if so do not continue
        if len(artworks) == 0:
            raise tvdb_attributenotfound
        #end if len(artworks) == 0
        
        if interactive:
            artworkCounter = 0
            print "\nList of available artwork"
            for artwork in artworks:
                print "%s. %s" % (artworkCounter, artwork)
                artworkCounter += 1
            #end for artwork
            
            #allow user to preview images
            print "Example of listing: 0 2 4"
            artworkPreviewRequestNumbers = raw_input("List Images to Preview: ")
            artworkPreviewRequests = artworkPreviewRequestNumbers.split()
            
            artworkPreviewUrls = []
            for artworkPreviewRequest in artworkPreviewRequests:
                artworkPreviewUrls.append(artworks[int(artworkPreviewRequest)])
            #end for artworkPreviewRequest
            openurl(artworkPreviewUrls)
            
            #ask user what artwork he wants to use
            artworkChoice = int(raw_input("Artwork to use: "))
        else:
            artworkChoice = 0
        #end if interactive
        
        artworkUrl = artworks[artworkChoice]
        
        (artworkUrl_base, artworkUrl_fileName) = os.path.split(artworkUrl)
        (artworkUrl_baseFileName, artworkUrl_fileNameExtension)=os.path.splitext(artworkUrl_fileName)
        
        artworkFileName = normalize('NFD', series.seriesName).encode("utf8") + " Season " + str(series.seasonNumber) + artworkUrl_fileNameExtension
        
        artworkFullFileName = cacheDir + "/" + artworkFileName
        
        if opts.verbose > 0:
            dlhook = verbosehook
            print "Downloading artwork: %s" % artworkFileName
        else:
            dlhook = ""
        #end if verbose
        
        urllib.urlretrieve(artworkUrl, artworkFullFileName, dlhook)
        
        series.artworkFileName = artworkFullFileName
    except tvdb_attributenotfound:
        # The attribute wasn't found, not critical
        if opts.verbose:
            sys.stderr.write("!! Non-Critical Show Error: %s not found for %s\n" % ("artwork", series.seriesName))
        #end if verbose
#end artwork


def verbosehook(blocks_read, block_size, total_size):
    if not blocks_read:
        global pbar
        pbar = progressbar.ProgressBar(maxval=total_size, widgets=[progressbar.Bar(), ' ', progressbar.Percentage()]).start()
        return
    amount_read = blocks_read * block_size
    if total_size < 0:
        # Unknown size
        print 'Read %d blocks' % blocks_read
    elif amount_read < total_size: 
        pbar.update(amount_read)
    elif total_size >= total_size:
        pbar.finish()
    return


def getShowSpecificInfo(opts, tvdb, seriesName, attribute):
    """docstring for getEpisodeSpecificInfo"""
    try:
        value = tvdb[seriesName][attribute]
        if not value:
            return ""
        #clean up string
        value =  value.replace('&quot;', "\\\"")
        return value
    except tvdb_error, errormsg:
        # Error communicating with thetvdb.com
        sys.stderr.write("!!!! Critical Show Error: Error contacting www.thetvdb.com:\n%s\n" % (errormsg))
        sys.exit(2)
    except tvdb_shownotfound:
        # No such series found.
        sys.stderr.write("!!!! Critical Show Error: Show %s not found\n" % (seriesName))
        sys.exit(2)
    except tvdb_seasonnotfound, errormsg:
        #the season name could not be found
        sys.stderr.write("!! Critical Show Error: The series name was not found for %s\n" % (seriesName))
        return 2
    except tvdb_attributenotfound, errormsg:
        # The attribute wasn't found, not critical
        if verbose:
            sys.stderr.write("!! Non-Critical Show Error: %s not found for %s\n" % (attribute, seriesName))
        #end if verbose
#end getEpisodeSpecificInfo


def getEpisodeSpecificInfo(opts, program, series, episodeNumber, attribute):
    """docstring for getEpisodeSpecificInfo"""
    try:
        value = program.tvdb[series.seriesName][series.seasonNumber][episodeNumber][attribute]
        if not value:
            return ""
        #clean up string
        value = value.replace('&quot;', "\\\"")
        value = value.replace('\"', "\\\"")
        value = value.replace('`', "'")
        return value
    except tvdb_episodenotfound:
        # The episode was not found wasn't found
        sys.stderr.write("!!!! Critical Episode Error: Episode name not found for %s - %02dx%02d\n" % (series.seriesName, series.seasonNumber, episodeNumber))
        sys.exit(2)
    except tvdb_error, errormsg:
        # Error communicating with thetvdb.com
        sys.stderr.write("!!!! Critical Episode Error: Error contacting www.thetvdb.com:\n%s\n" % (errormsg))
        sys.exit(2)
    except tvdb_attributenotfound:
        # The attribute wasn't found, not critical
        if verbose:
            sys.stderr.write("!! Non-Critical Episode Error: %s not found for %02dx%02d\n" % (attribute, series.seasonNumber, episodeNumber))
        #end if verbose
        return ""
#end getEpisodeSpecificInfo


def tvtags(opts, fullPath):
    
    if not whichBin("MP4Tagger"):
        print "MP4Tagger tools not found"
        sys.exit(0)
    
    (filePath, fileName) = os.path.split(fullPath)
    
    if not os.path.isfile(fullPath):
        sys.stderr.write("!!!! Critical Error:  file \"%s\" does not exist\n" % (fileName))
        sys.exit(2)
    #end if not os.path.isfile
    
    if len(filePath) == 0:
        filePath = "."
    
    m1 = re.match('(.+).[sS]([0-9]+)[eE]([0-9]+).+', fileName)
    m2 = re.match('(.+).([0-9]+)[xX]([0-9]+).+', fileName)
    if m1:
        (series, seasonNumber, episodeNumber) = m1.groups()
    elif m2:
        (series, seasonNumber, episodeNumber) = m2.groups()
    else :
        sys.stderr.write("!!!! Critical Error: file name \"%s\" is of incorrect format\nExample of file name: invasion.s01e02.hr.hdtv.xvid-nbs.m4v\n" % (fileName))
        sys.exit(2)
    
    if opts.verbose > 1:
        print ("analyse => serie: %s saison: %s episode: %s" % (series, seasonNumber, episodeNumber))
    #end if debug
    
    program = Program(opts, filePath, fileName)
    
    series = Series(opts, program, series, seasonNumber)
    
    #request user to select artwork
    artworkFileName = artwork(opts, opts.interactive, program, series)
    
    #check if the image we have needed resizing/dpi changed -> use this new temp file that was created for all the other episodes
    (imageFile, imageExtension) = os.path.splitext(series.artworkFileName)
    if series.artworkFileName.count("-resized-") == 0:
        for imageFileName in glob.glob("*" + imageExtension):
            if imageFileName.count("-resized-"):
                series.artworkFileName = imageFileName
                if opts.verbose:
                    print "Using resized artwork file \"%s\"" % imageFileName
                #end if opts.verbose
                break
            #end if imageFileName.count
        #end for imageFileName
    #end if series.artworkFileName.count
    
    #create an episode which will populate it's fields using data from thetvdb
    episode = Episode(opts.verbose, program, series, seasonNumber, episodeNumber)
    
    if opts.rename:
        #fix the filename
        correctFileName(opts.verbose, program, series, episode)
    #end if opts.rename
    
    tagFile(opts, program, series, episode)
#end tvtags


def main():
    "docstring for tvtags"
    parser = OptionParser(usage="%prog [options] <path to video file(s)>\n%prog -h for full list of options")
    
    parser.add_option(  "-b", "--batch", action="store_false", dest="interactive", help="selects first search result, requires no human intervention once launched [default]")
    parser.add_option(  "-i", "--interactive", action="store_true", dest="interactive", help="interactivly select correct show from search results")
    parser.add_option(  "-d", "--debug", action="store_const", const=2, dest="verbose", help="shows all debugging info")
    parser.add_option(  "-v", "--verbose", action="store_const", const=1, dest="verbose", help="Will provide some feedback [default]")
    parser.add_option(  "-q", "--quiet", action="store_const", const=0, dest="verbose", help="For ninja-like processing")
    parser.add_option(  "-n", "--renaming", action="store_true", dest="rename", help="enable cleaning name")
    parser.set_defaults( interactive=False, verbose=1, rename=False  )
    
    opts, args = parser.parse_args()
    
    if len(args) == 0:
        parser.error("No video file supplied")
    #end if len(args)
    
    for fullPath in args:
        tvtags(opts, fullPath)
#end main

if __name__ == '__main__':
    sys.exit(main())
#end if __name__


#!/usr/bin/env python

__version__ = "0.7.0"

from setuptools import setup, find_packages
setup(
name = 'tvtags',
version = __version__,
author ='tuxtof',
author_email ='dev@geo6.net',
description ='Automagicly media tagger.',
url ='http://github.com/tuxtof/tvtags',
license ='GPLv2',
long_description ='tvtags use thetvdb database to automaticly tag media file.',

install_requires = ['progressbar>=2.2'],

py_modules = ['tvtags'],
packages = find_packages(),


entry_points = {
    'console_scripts': [
        'tvtags = tvtags:main',
    ],
},

)
